#include "MacPython.h"

#include <stdio.h>
#include <string.h>
#include <dlfcn.h>
#include <stdlib.h>
#include <wchar.h>

#define ROOTPATH "/Library/Frameworks/"
#define LROOTPATH L"/Library/Frameworks/"

#define PYTHONLIB_PATH ROOTPATH "Python.framework/Versions/3.1/Python"

#define PyUnicode_AsEncodedString PyUnicodeUCS2_AsEncodedString

typedef long Py_ssize_t;

typedef struct _object
{
	Py_ssize_t ob_refcnt;
	struct _typeobject *ob_type;
} PyObject;

typedef struct
{
	void (* Py_Initialize)(void);
	void (* Py_Finalize)(void);
	void (* PySys_SetArgv)(int, wchar_t **);
	PyObject * (*PyImport_ExecCodeModule)(char *, PyObject *);
	PyObject *(*PyMarshal_ReadObjectFromString)(char *, Py_ssize_t);
	void (*Py_IncRef)(PyObject *);
	void (*Py_DecRef)(PyObject *);
	PyObject *(*Py_BuildValue)(const char *, ...);
	const char *(*Py_GetVersion)(void);
	const char *(*Py_GetPlatform)(void);
	PyObject *(*PyModule_GetDict)(PyObject *);
	PyObject *(*PyDict_GetItemString)(PyObject *, const char *);
	PyObject * (*PyObject_CallFunction)(PyObject *, char *format, ...);
	PyObject*(*PyUnicodeUCS2_AsEncodedString)(PyObject *, const char *, const char *);
	PyObject * (*PyBytes_FromString)(const char *);
	char *(*PyBytes_AsString)(PyObject *);
	char *(*PyString_AsString)(PyObject *);
	
	wchar_t *(*Py_GetProgramFullPath)(void);
	wchar_t *(*Py_GetPrefix)(void);
	wchar_t *(*Py_GetExecPrefix)(void);
	wchar_t *(*Py_GetPath)(void);
	int (*PyNumber_Check)(PyObject *);
	PyObject * (*PyTuple_New)(Py_ssize_t);
	int (*PyTuple_SetItem)(PyObject *, Py_ssize_t, PyObject *);
	void (*PyErr_Clear)(void);
	
	PyObject *(*PyImport_ImportModule)(const char *);
	PyObject *(*PyObject_GetAttrString)(PyObject*, const char *);
	int (*PySys_SetObject)(const char *, PyObject *);
	PyObject *(*PyObject_CallObject)(PyObject *, PyObject *);
	PyObject* (*PyObject_Type)(PyObject *);
	int (*PyObject_Print)(PyObject *, FILE *, int);
	PyObject *(*PyErr_Occurred)(void);
	void (*PyErr_Print)(void);	
	
} python_functions_t;

python_functions_t python = {0};

PyObject *GetValueFunction = NULL;
PyObject *StringIOObject = NULL;
PyObject *ioModule = NULL;
PyObject *StringIOFunction = NULL;
bool logging = false;


void *loadPythonFramework(void)
{
	void *pythonLib = dlopen(PYTHONLIB_PATH, RTLD_NOW);
	if (!pythonLib)
		return NULL;
	
	if ((python.Py_Initialize = (void (*)(void)) dlsym( pythonLib, "Py_Initialize" )) == NULL) { return NULL; }
	if ((python.Py_Finalize = (void (*)(void)) dlsym( pythonLib, "Py_Finalize" )) == NULL) { return NULL; }
	if ((python.PySys_SetArgv = (void (*)(int, wchar_t **)) dlsym( pythonLib, "PySys_SetArgv" )) == NULL) { return NULL; }
	if ((python.PyImport_ExecCodeModule = (PyObject* (*)(char *, PyObject*)) dlsym( pythonLib, "PyImport_ExecCodeModule" )) == NULL) { return NULL; }
	if ((python.PyMarshal_ReadObjectFromString = (PyObject* (*)(char *, Py_ssize_t)) dlsym( pythonLib, "PyMarshal_ReadObjectFromString" )) == NULL) { return NULL; }
	if ((python.Py_IncRef = (void (*)(PyObject *)) dlsym( pythonLib, "Py_IncRef" )) == NULL) { return NULL; }
	if ((python.Py_DecRef = (void (*)(PyObject *)) dlsym( pythonLib, "Py_DecRef" )) == NULL) { return NULL; }
	if ((python.Py_BuildValue = (PyObject* (*)(const char *, ...)) dlsym( pythonLib, "Py_BuildValue" )) == NULL) { return NULL; }
	if ((python.Py_GetVersion = (const char* (*)(void)) dlsym( pythonLib, "Py_GetVersion" )) == NULL) { return NULL; }
	if ((python.Py_GetPlatform = (const char* (*)(void)) dlsym( pythonLib, "Py_GetPlatform" )) == NULL) { return NULL; }
	if ((python.PyModule_GetDict = (PyObject* (*)(PyObject *)) dlsym( pythonLib, "PyModule_GetDict" )) == NULL) { return NULL; }
	if ((python.PyDict_GetItemString = (PyObject* (*)(PyObject *, const char *)) dlsym( pythonLib, "PyDict_GetItemString" )) == NULL) { return NULL; }
	if ((python.PyObject_CallFunction = (PyObject * (*)(PyObject *, char *format, ...)) dlsym( pythonLib, "PyObject_CallFunction" )) == NULL) { return NULL; }
	if ((python.PyUnicodeUCS2_AsEncodedString = (PyObject* (*)(PyObject *, const char *, const char *)) dlsym( pythonLib, "PyUnicodeUCS2_AsEncodedString" )) == NULL) { return NULL; }
	if ((python.PyBytes_FromString = (PyObject* (*)(const char *)) dlsym( pythonLib, "PyBytes_FromString" )) == NULL) { return NULL; }
	if ((python.PyBytes_AsString = (char* (*)(PyObject *)) dlsym( pythonLib, "PyBytes_AsString" )) == NULL) { return NULL; }
	if ((python.Py_GetPath = (wchar_t* (*)(void)) dlsym( pythonLib, "Py_GetPath" )) == NULL) { return NULL; }
	if ((python.Py_GetPrefix = (wchar_t *(*)(void)) dlsym( pythonLib, "Py_GetPrefix" )) == NULL) { return NULL; }
	if ((python.Py_GetExecPrefix = (wchar_t *(*)(void)) dlsym( pythonLib, "Py_GetExecPrefix" )) == NULL) { return NULL; }
	if ((python.Py_GetProgramFullPath = (wchar_t *(*)(void)) dlsym( pythonLib, "Py_GetProgramFullPath" )) == NULL) { return NULL; }
	if ((python.PyNumber_Check = (int (*)(PyObject *)) dlsym( pythonLib, "PyNumber_Check" )) == NULL) { return NULL; }
	if ((python.PyTuple_New = (PyObject * (*)(Py_ssize_t)) dlsym( pythonLib, "PyTuple_New" )) == NULL) { return NULL; }
	if ((python.PyTuple_SetItem = (int (*)(PyObject *, Py_ssize_t, PyObject *)) dlsym( pythonLib, "PyTuple_SetItem" )) == NULL) { return NULL; }
	if ((python.PyErr_Clear = (void (*)(void)) dlsym( pythonLib, "PyErr_Clear" )) == NULL) { return NULL; }
	
	if ((python.PyImport_ImportModule = (PyObject *(*)(const char *)) dlsym( pythonLib, "PyImport_ImportModule" )) == NULL) { return NULL; }
	if ((python.PyObject_GetAttrString = (PyObject *(*)(PyObject*, const char *)) dlsym( pythonLib, "PyObject_GetAttrString" )) == NULL) { return NULL; }
	if ((python.PySys_SetObject = (int (*)(const char *, PyObject *)) dlsym( pythonLib, "PySys_SetObject" )) == NULL) { return NULL; }
	if ((python.PyObject_CallObject = (PyObject *(*)(PyObject *, PyObject *)) dlsym( pythonLib, "PyObject_CallObject" )) == NULL) { return NULL; }
	if ((python.PyObject_Type = (PyObject* (*)(PyObject *)) dlsym( pythonLib, "PyObject_Type" )) == NULL) { return NULL; }
	if ((python.PyObject_Print = (int (*)(PyObject *, FILE *, int)) dlsym( pythonLib, "PyObject_Print" )) == NULL) { return NULL; }
	if ((python.PyErr_Occurred = (PyObject *(*)(void)) dlsym( pythonLib, "PyErr_Occurred" )) == NULL) { return NULL; }
	if ((python.PyErr_Print = (void (*)(void)) dlsym( pythonLib, "PyErr_Print" )) == NULL) { return NULL; }	
	
	return pythonLib;
}

void unloadPythonFramework( void *pythonLib )
{
	dlclose(pythonLib);
}

void initializePython(void)
{
	python.Py_Initialize();
	
	ioModule = python.PyImport_ImportModule("io");
	if (!ioModule) goto EXIT;
	
	StringIOFunction = python.PyObject_GetAttrString(ioModule, "StringIO");
	if (!StringIOFunction) goto EXIT;
	
	StringIOObject = python.PyObject_CallObject(StringIOFunction, NULL);
	if (!StringIOObject) goto EXIT;
	
	if ( python.PySys_SetObject("stderr", StringIOObject) != 0 ) goto EXIT;
	
	GetValueFunction = python.PyObject_GetAttrString(StringIOObject, "getvalue");
	if (!GetValueFunction) goto EXIT;
	
	return;
	
EXIT:
	if (ioModule) python.Py_DecRef(ioModule);
	if (StringIOFunction) python.Py_DecRef(StringIOFunction);
	if (StringIOObject) python.Py_DecRef(StringIOObject);
	if (GetValueFunction) python.Py_DecRef(GetValueFunction);
}

void finalizePython(void)
{
	PyObject *result = NULL;
	PyObject* resultString = NULL;
	char *returnString = NULL;
	
	if ( python.PyErr_Occurred() )
	{
		python.PyErr_Print(); // This will clear the error and print it to stderr
	}
	
	result = python.PyObject_CallObject(GetValueFunction, NULL);
	if (!result) goto EXIT;
	
	resultString = python.PyUnicode_AsEncodedString(result, "utf8", NULL);
	if (!resultString) goto EXIT;
	
	returnString = python.PyBytes_AsString(resultString);
	printf( "%s\n", returnString );

	
EXIT:
	if (result) python.Py_DecRef(result);
	if (resultString) python.Py_DecRef(resultString);
	if (ioModule) python.Py_DecRef(ioModule);
	if (StringIOFunction) python.Py_DecRef(StringIOFunction);
	if (StringIOObject) python.Py_DecRef(StringIOObject);
	if (GetValueFunction) python.Py_DecRef(GetValueFunction);
	
	python.Py_Finalize();
}

char *executeModuleFunction(const char *module, size_t moduleSize, const char *moduleName, const char *stringIn, const char *functionName)
{
	const size_t PYC_MAGIGBYTES_SIZE = 8;
	char *resultStr = NULL;
	wchar_t *params[] = { NULL, NULL };
	wchar_t *wName = NULL;
	PyObject *codeObject = NULL, *newModule = NULL, *globalDict = NULL, *function = NULL, *result = NULL;
	size_t size = 0, charsConverted = 0;
	
	if (moduleSize <= PYC_MAGIGBYTES_SIZE)
	{
		return NULL;
	}
	
	size = strlen(moduleName) + 1;
	wName = (wchar_t *)malloc(size * sizeof(wchar_t));
	if (wName == NULL)
	{
		return NULL;
	}
	
	charsConverted = mbstowcs(wName, moduleName, size);
	
	if (charsConverted < (size - 1) || charsConverted == (size_t)-1)
	{
		free(wName);
		return NULL;
	}
	
	params[0] = wName;
	
	python.PySys_SetArgv( 1, params );	
	
	free(wName);

	codeObject = python.PyMarshal_ReadObjectFromString((char*)module + PYC_MAGIGBYTES_SIZE, moduleSize - PYC_MAGIGBYTES_SIZE);
	if (codeObject == NULL)
	{
		return NULL;
	}
	
	if ((newModule = python.PyImport_ExecCodeModule((char*)moduleName, codeObject)) != NULL &&
    	(globalDict = python.PyModule_GetDict(newModule)) != NULL &&
    	(function = python.PyDict_GetItemString(globalDict, functionName)) != NULL &&
		(result = python.PyObject_CallFunction(function, "y", stringIn)) != NULL)
    {
		if (python.PyNumber_Check(result))
		{
			/* Add some logging when the meaning of the return codes is clarified. */
		}
		else
		{
	        char* xmlStr = python.PyBytes_AsString(result);
			
			if (xmlStr != NULL)
			{
				resultStr = malloc(strlen(xmlStr) + 1);
				if (resultStr != NULL)
				{
					strcpy(resultStr, xmlStr);
				}
				else
				{
					return NULL;
				}
			}
			else
			{
				return NULL;
			}
        }
		python.Py_DecRef(result);
    }
	
	if (newModule != NULL)
		python.Py_DecRef(newModule);
	python.Py_DecRef(codeObject);
	
	return resultStr;
}

