
#include "MacPython.h"

#include <stdlib.h>
#include <stdio.h>
#include <sys/resource.h>
#include <string.h>
#include <stdbool.h>


short CallPython(const char* stringIn, char** stringOut, const char* pythonModule, size_t moduleSize,
                  const char* moduleName, const char* functionName)
{
	static void *handle = NULL;
	short result = 0;
	
	if (stringIn == NULL || stringOut == NULL || pythonModule == NULL || moduleName == NULL || functionName == NULL)
		return 1;
	
	if (handle == NULL) {
		handle = loadPythonFramework();
	}
	if (handle == NULL)
		return 2;

	initializePython();
	
	*stringOut = executeModuleFunction(pythonModule, moduleSize, moduleName, stringIn, functionName);
	if (*stringOut == NULL)
		result = 3;

	finalizePython();
	
	// Mac OS X : unloading and reloading the Python framework consumes more memory than leaving it open.
	// unloadPythonFramework( handle ); handle == NULL;
	
	return result;
}

char *ReadContentOfFile( char* path, unsigned long *size ) {
	FILE *file;
	char *content;
	unsigned long fileSize;
	
	if( ( file = fopen( path, "r")) == NULL) {
		puts("Cannot open the file");
		exit( 0 );
	}
	
	fseek(file, 0, SEEK_END);
	fileSize = ftell(file);
	
	fseek(file, 0, SEEK_SET);
	content = (char *)malloc(fileSize+1);
	
	if (fileSize != fread(content, sizeof(char), fileSize, file)) {
		free(content);
		return NULL;
	}
	
	fclose(file);
	content[fileSize] = 0;
	
	if (size) *size = fileSize;
	return content;
}


int main (int argc, const char * argv[]) {
	unsigned long size;
	char *content = ReadContentOfFile( "test.pyc", &size );
	char* stringOut;
		
	for (int x = 0; x <= 1000; x++) {
		printf("x = %d\n", x);
		CallPython( "ignored string", &stringOut, content, size, "test", "mainFunction");
		free(stringOut);
	}

	free(content);
		
	return 0;
}

